import os
import zipfile
import tempfile
import shutil
import re
from . import utils
from langchain.text_splitter import RecursiveCharacterTextSplitter

# Purpose: Get the absolute path of the calling script and join it with the target file


def this_dir(file):
    return os.path.dirname(os.path.realpath(file))


# Remove most non-alphanumeric characters from a filename


def clean_paths(file_path_or_paths):
    if (isinstance(file_path_or_paths, list)):
        paths = []
        for file_path in file_path_or_paths:
            paths.append(clean_paths(file_path))
        return paths
    return file_path_or_paths.replace("/", "_").replace("\\", "_").replace(" ", "_").replace(":", "").replace(",", "").replace("'", "").replace('"', '').lower()


def flatten_abs_paths(dir_or_file):
    abs_file_paths = []
    if os.path.isfile(dir_or_file):
        abs_file_paths.append(os.path.abspath(dir_or_file))
        return abs_file_paths

    for dirpath, _, filenames in os.walk(dir_or_file):
        for filename in filenames:
            file_path = os.path.join(dirpath, filename)
            abs_file_paths.append(os.path.abspath(file_path))
    return abs_file_paths


def unzip_and_remove(zip_path, extract_to):
    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        zip_ref.extractall(extract_to)
    os.remove(zip_path)  # remove the .zip file after extraction


def _unzip_temp(temp_dir):
    for root, dirs, files in os.walk(temp_dir):
        for filename in files:
            if filename.endswith(".zip"):
                zip_file_path = os.path.join(root, filename)
                unzip_and_remove(zip_path=zip_file_path, extract_to=root)
                _unzip_temp(root)  # recursive call to handle nested zip files


def unzip_all(file_or_dir, dest_dir=None):
    # Copy all into a temp dir
    temp_dir = tempfile.mkdtemp()
    shutil.copytree(file_or_dir, temp_dir, dirs_exist_ok=True)

    # Recursively unzip zipped files
    _unzip_temp(temp_dir)

    # Move all files to dest_dir (if exists)
    if dest_dir:
        shutil.copytree(temp_dir, dest_dir, dirs_exist_ok=True)
        shutil.rmtree(temp_dir)
        return dest_dir

    return temp_dir

# Purpose: Create a directory using the url or path in ~/.locallm/chunks directory


def get_chunk_dir(chunk_dir, path_or_url):
    if (utils.is_url(path_or_url)):
        path_or_url = path_or_url.replace(
            "://", "_").replace("/", "_").replace(".", "_")
    else:
        path_or_url = os.path.basename(path_or_url).split('.')[0]
    chunk_name = re.sub(r'^_+', '', path_or_url)
    return os.path.join(chunk_dir, chunk_name)

# Purpose: Create a chunk id using sha256 of its context


def create_chunk_id(text):
    import hashlib
    if isinstance(text, bytes):
        return hashlib.sha256(text).hexdigest()
    if isinstance(text, str):
        return hashlib.sha256(text.encode('utf-8')).hexdigest()
    raise Exception("content must be bytes or text")

# Purpose: Split text using LangChain's recursive text splitter into chunks. Optionally save the chunks in the chunks_dir


def split_chunks(text, chunks_dir=None, chunk_size=1000, chunk_overlap=100):

    # Remove existing chunks
    if (chunks_dir != None):
        if os.path.exists(chunks_dir):
            shutil.rmtree(chunks_dir)
        os.makedirs(chunks_dir)

    splitter = RecursiveCharacterTextSplitter(
        chunk_size=chunk_size,        # approx 512 tokens
        chunk_overlap=chunk_overlap,     # 10% overlap
        length_function=len,
        is_separator_regex=False
    )
    chunks = splitter.create_documents([text])
    for chunk in chunks:
        chunk_id = create_chunk_id(chunk.page_content)
        chunk.metadata = {
            "chunk_id": chunk_id,
            "chunk_size": len(chunk.page_content)
        }
        if (chunks_dir != None):
            chunk_fname = os.path.join(chunks_dir, chunk_id)
            with open(chunk_fname, 'w') as f:
                f.write(chunk.page_content)
    return chunks



import uuid
from fastapi.responses import JSONResponse

class ApiException(Exception):
    def __init__(self, status_code, code, message, url):
        self.status_code = status_code
        self.code = code
        self.url = url
        super().__init__(message)
    pass

class ErrorResponse(JSONResponse):
    def __init__(self, status_code, code, message):
        super().__init__(
            content={
                "type": "error",
                "code": code,
                "message": message
            },
            status_code=status_code
        )

class ContextLengthExceededError(ErrorResponse):
    def __init__(self):
        super().__init__(
            status_code=400,
            code="context_length_exceeded",
            message="The message has exceeded the model's context length."
        )

class ModelServiceMismatchError(ErrorResponse):
    def __init__(self):
        super().__init__(
            status_code=400,
            code="model_service_mismatch",
            message="The model does not correspond to this service."
        )

class InternalError(ErrorResponse):
    def __init__(self, error_id):
        super().__init__(
            status_code=500,
            code=error_id,
            message="An unexpected error occurred. Please try again later."
        )



import websockets
import json


class StatusListener:

    def __init__(self, uri):
        self.uri = uri
        self.cancellation_token = None

    async def listen(self, callback=None):
        async with websockets.connect(self.uri) as websocket:
            print(f"Connected to {self.uri}")
            try:
                while self.cancellation_token is None:
                    message = await websocket.recv()
                    if callback:
                        callback(message)
                    print(f"Received status update: {message}")

                    if message == "<stop>":
                        self.cancellation_token = message

            except websockets.exceptions.ConnectionClosed as e:
                print(f"Connection closed: {e}")

    def stop(self):
        self.cancellation_token = "<stop>"
        print("Stopping listener")
        return self.cancellation_token
    
        


import os
import requests
import json
import gai.common.ConfigHelper as ConfigHelper
from fastapi import WebSocketDisconnect
from gai.common.http_utils import http_post, http_delete,http_get,http_delete_async
from gai.common.logging import getLogger
from gai.common.errors import ApiException
logger = getLogger(__name__)
import asyncio
from gai.lib.ClientBase import ClientBase

class RAGClient(ClientBase):

    def __init__(self,config_path=None):
        super().__init__(config_path)
        self.base_url = os.path.join(
            self.config["gai_url"], 
            self.config["generators"]["rag"]["url"].lstrip('/'))
        logger.debug(f'base_url={self.base_url}')

    # Provides an updater to get chunk indexing status
    # NOTE: The update is only relevant if this library is used in a FastAPI application with a websocket connection
    async def index_file_async(self, collection_name, file_path, metadata={"source": "unknown"}, progress_updater=None):
        url=os.path.join(self.base_url,"index-file")

        # We will assume file ending with *.pdf to be PDF but this check should be done before the call.
        mode = 'rb' if file_path.endswith('.pdf') else 'r'
        with open(file_path, mode) as f:
            files = {
                "file": (os.path.basename(file_path), f, "application/pdf"),
                "metadata": (None, json.dumps(metadata), "application/json"),
                "collection_name": (None, collection_name, "text/plain")
            }
            response = http_post(url=url, files=files)

        # Callback for progress update (returns a number between 0 and 100)
        if progress_updater:
            # Exception should not disrupt the indexing process
            try:
                # progress = int((i + 1) / len(chunks) * 100)
                progress = 100
                await progress_updater(progress)
                logger.debug(
                    f"RAGClient: progress={progress}")
                # await send_progress(websocket, progress)
            except WebSocketDisconnect as e:
                if e.code == 1000:
                    # Normal closure, perhaps log it as info and continue gracefully
                    logger.info(
                        f"RAGClient: WebSocket closed normally with code {e.code}")
                    pass
                else:
                    # Handle other codes as actual errors
                    logger.error(
                        f"RAGClient: WebSocket disconnected with error code {e.code}")
                    pass
            except Exception as e:
                logger.error(
                    f"RetrievalGeneration.index_async: Update websocket progress failed. Error={str(e)}")
                pass

            return response

    # synchronous version of index_file_async
    def index_file(self, collection_name, file_path, metadata={"source": "unknown"}, progress_updater=None):
        url = os.path.join(self.base_url,"index-file")

        # We will assume file ending with *.pdf to be PDF but this check should be done before the call.
        mode = 'rb' if file_path.endswith('.pdf') else 'r'
        with open(file_path, mode) as f:
            files = {
                "file": (os.path.basename(file_path), f.read(), "application/pdf"),
                "metadata": (None, json.dumps(metadata), "application/json"),
                "collection_name": (None, collection_name, "text/plain")
            }
            response = http_post(url=url, files=files)

        # Callback for progress update (returns a number between 0 and 100)
        if progress_updater:
            # Exception should not disrupt the indexing process
            try:
                # progress = int((i + 1) / len(chunks) * 100)
                progress = 100
                t = asyncio.create_task(progress_updater(progress))
                asyncio.get_event_loop().run_until_complete(t)
                logger.debug(
                    f"RAGClient: progress={progress}")
                # await send_progress(websocket, progress)
            except WebSocketDisconnect as e:
                if e.code == 1000:
                    # Normal closure, perhaps log it as info and continue gracefully
                    logger.info(
                        f"RAGClient: WebSocket closed normally with code {e.code}")
                    pass
                else:
                    # Handle other codes as actual errors
                    logger.error(
                        f"RAGClient: WebSocket disconnected with error code {e.code}")
                    pass
            except Exception as e:
                logger.error(
                    f"RetrievalGeneration.index_async: Update websocket progress failed. Error={str(e)}")
                pass
        
        # {document_id: "document_id"}
        return json.loads(response.text)

    def retrieve(self, collection_name, query_texts, n_results=None):
        url = os.path.join(self.base_url,"retrieve")
        data = {
            "collection_name": collection_name,
            "query_texts": query_texts
        }
        if n_results:
            data["n_results"] = n_results

        response = http_post(url, data=data)
        return response

    # Database Management

    def delete_collection(self, collection_name):
        url = os.path.join(self.base_url,"collection",collection_name)
        logger.info(f"RAGClient.delete_collection: Deleting collection {url}")
        try:
            response = http_delete(url)
        except ApiException as e:
            if e.code == 'collection_not_found':
                return {"count":0}
            logger.error(e)
            raise e
        return json.loads(response.text)

    async def delete_collection_async(self, collection_name):
        url = os.path.join(self.base_url,"collection",collection_name)
        logger.info(f"RAGClient.delete_collection: Deleting collection {url}")
        try:
            response = await http_delete_async(url)
        except ApiException as e:
            if e.code == 'collection_not_found':
                return {"count":0}
            logger.error(e)
            raise e
        return json.loads(response.text)

    def list_collections(self):
        url = os.path.join(self.base_url,"collections")
        response = http_get(url)
        return json.loads(response.text)

    def list_documents(self,collection_name):
        url = os.path.join(self.base_url,"collection",collection_name)
        response = http_get(url)
        return json.loads(response.text)
    
    def get_document(self,doc_id):
        url = os.path.join(self.base_url,"document",doc_id)
        response = http_get(url)
        return json.loads(response.text)

    def delete_document(self,doc_id):
        url = os.path.join(self.base_url,"document",doc_id)
        response = http_delete(url)
        return json.loads(response.text)





from gai.lib.ttt.ChunkWrapper import ChunkWrapper
from gai.lib.ttt.OpenAIChunkWrapper import OpenAIChunkWrapper
from gai.lib.ttt.AnthropicChunkWrapper import AnthropicChunkWrapper
import gai.common.ConfigHelper as ConfigHelper
from gai.common.http_utils import http_post
from gai.common.generators_utils import chat_string_to_list, chat_list_to_string
from gai.common.errors import ApiException
from gai.common.logging import getLogger
logger = getLogger(__name__)
import json,os
from gai.lib.ClientBase import ClientBase

class TTTClient(ClientBase):

    def __init__(self, config_path=None):
        super().__init__(config_path)
        logger.debug(f'base_url={self.base_url}')

    def api(self, generator="mistral7b-exllama", messages=None, stream=True, **generator_params):
        if not messages:
            raise Exception("Messages not provided")

        if isinstance(messages, str):
            messages = chat_string_to_list(messages)

        if not generator:
            generator = self.config["default_generator"]

        data = {
            "model": generator,
            "messages": messages,
            "stream": stream,
            **self.config["generators"][generator]["default"],
            **generator_params
        }

        def streamer(response):
            for chunk in response.iter_lines():
                yield ChunkWrapper(chunk)

        try:
            url = self._gen_url(generator)
            response = http_post(url, data)
        except ApiException as he:

            # Switch to Mistral7b 128k context size
            if he.code == "context_length_exceeded":
                try:
                    generator = "mistral7b_128k-exllama"
                    data["model"] = generator
                    url = self._gen_url(generator)
                    response = http_post(url, data)
                except Exception as e:
                    raise e
            else:
                raise he
        except Exception as e:
            raise e

        if not stream:
            if response.json()["choices"][0]["message"]["tool_calls"]:
                response.decode = lambda: {
                    "type":"function",
                    "name": response.json()["choices"][0]["message"]["tool_calls"][0]["function"]["name"],
                    "arguments": response.json()["choices"][0]["message"]["tool_calls"][0]["function"]["arguments"],
                }
            elif response.json()["choices"][0]["message"]["content"]:
                response.decode = lambda: {
                    "type": "content",
                    "content": response.json()["choices"][0]["message"]["content"]
                }
            return response
        return streamer(response)

    def __call__(self, generator=None, messages=None, stream=True, **generator_params):
        if generator == "gpt-4":
            return self.gpt_4(messages=messages, stream=stream, **generator_params)
        elif generator == "claude2-100k":
            return self.claude_2(messages=messages, stream=stream, **generator_params)
        return self.api(generator, messages, stream, **generator_params)

    def gpt_4(self, messages=None, stream=True, **generator_params):
        import os
        import openai
        from openai import OpenAI
        from dotenv import load_dotenv
        load_dotenv()
        if not os.environ.get("OPENAI_API_KEY"):
            raise Exception(
                "OPENAI_API_KEY not found in environment variables")
        openai.api_key = os.environ["OPENAI_API_KEY"]
        client = OpenAI()

        if not messages:
            raise Exception("Messages not provided")

        def streamer(response):
            for chunk in response:
                yield OpenAIChunkWrapper(chunk)

        model = "gpt-4"
        if isinstance(messages, str):
            messages = chat_string_to_list(messages)
        response = client.chat.completions.create(
            model=model,
            messages=messages,
            stream=stream,
            **generator_params
        )

        if not stream:
            if response.choices[0].message.tool_calls:
                response.decode = lambda: {
                    "type":"function",
                    "name": response.choices[0].message.tool_calls[0].function.name,
                    "arguments": response.choices[0].message.tool_calls[0].function.arguments,
                }
            elif response.choices[0].message.content:
                response.decode = lambda: {
                    "type": "content",
                    "content": response.choices[0].message.content
                }
            return response


        if not stream:
            response.decode = lambda: response.choices[
                0].message.content if response.choices[0].message.content else ""
            return response
        return streamer(response)

    def claude_2(self, messages=None, stream=True, **generator_params):
        import os
        from anthropic import Anthropic
        from dotenv import load_dotenv
        load_dotenv()
        if not os.environ.get("ANTHROPIC_API_KEY"):
            raise Exception(
                "ANTHROPIC_API_KEY not found in environment variables")
        client = Anthropic()

        if not messages:
            raise Exception("Messages not provided")

        def streamer(response):
            for chunk in response:
                yield AnthropicChunkWrapper(chunk)

        model = "claude-2"
        message = messages
        if isinstance(messages, list):
            message = chat_list_to_string(messages)
        prompt_template = "\n\nHuman: {message}\n\nAssistant:"
        messages = prompt_template.format(message=message)
        response = client.completions.create(
            model=model,
            prompt=messages,
            stream=stream,
            **generator_params
        )

        if not stream:
            response.decode = lambda: response.completion
            return response
        return streamer(response)



from gai.lib.ttt.TTTClient import TTTClient
from gai.lib.STTClient import STTClient
from gai.lib.TTSClient import TTSClient
from gai.lib.ITTClient import ITTClient
from gai.lib.RAGClient import RAGClient
from PIL import Image
import base64
from io import BytesIO

class GGG:
    client = None

    def __init__(self,config_path=None):
        self.config_path = config_path

    def __call__(self, category, **model_params):
        if category.lower() == "ttt":
            self.client = TTTClient(self.config_path)
            return self.client(**model_params)
        elif category.lower() == "ttt-long":
            self.client = TTTClient(self.config_path)
            generator = model_params.pop("generator", None)
            generator = "mistral7b_128k-exllama"
            return self.client(generator, **model_params)
        elif category.lower() == "stt":
            self.client = STTClient(self.config_path)

            if "file_path" in model_params:
                file_path = model_params.pop("file_path", None)
                return self.client(file=open(file_path, "rb"), **model_params)

            if "file" in model_params:
                file = model_params.pop("file", None)
                return self.client(file=file, **model_params)

            return self.client(**model_params)
        elif category.lower() == "tts":
            self.client = TTSClient(self.config_path)
            return self.client(**model_params)
        elif category.lower() == "itt":
            self.client = ITTClient(self.config_path)
            messages = None

            if "messages" in model_params:
                messages = model_params.pop("messages", None)
                return self.client(messages=messages, **model_params)

            text = "Describe the image"
            if "text" in model_params:
                text = model_params.pop("text", None)

            def process_image(image, text):
                image_format = image.format
                buffered = BytesIO()
                image.save(buffered, format=image_format)
                image_base64 = base64.b64encode(buffered.getvalue()).decode()
                return [
                    {"role": "user", "content": [
                        {"type": "text", "text": text},
                        {"type": "image_url", "image_url": {
                            "url": f"data:image/{image_format};base64,"+image_base64
                        }
                        }
                    ]}
                ]

            if "file_path" in model_params:
                file_path = model_params.pop("file_path", None)
                with open(file_path, "rb") as f:
                    image = Image.open(f)
                    messages = process_image(image, text)
                    return self.client(messages=messages, **model_params)

            if "file" in model_params:
                file = model_params.pop("file", None)
                image = Image.open(file)
                messages = process_image(image, text)
                return self.client(messages=messages, **model_params)

            if "image" in model_params:
                image = model_params.pop("image", None)
                messages = process_image(image, text)
                return self.client(messages=messages, **model_params)

            if "image_url" in model_params:
                image_url = model_params.pop("image_url", None)
                messages = [
                    {"role": "user", "content": [
                        {"type": "text", "text": text},
                        {"type": "image_url", "image_url": image_url}
                    ]}
                ]
                return self.client(messages=messages, **model_params)
        elif category.lower() == "index":
            self.client = RAGClient(self.config_path)
            return self.client.index_file(**model_params)
        elif category.lower() == "retrieve":
            self.client = RAGClient(self.config_path)
            return self.client.retrieve(**model_params)
        else:
            raise Exception(f"Unknown category: {category}")



# Copied from https://stackoverflow.com/a/1336640

import logging
import platform
import os

logging.basicConfig(
    format='%(asctime)s %(levelname)s %(name)s:%(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
)

def add_coloring_to_emit_windows(fn):
    # add methods we need to the class
    def _out_handle(self):
        import ctypes
        return ctypes.windll.kernel32.GetStdHandle(self.STD_OUTPUT_HANDLE)
    out_handle = property(_out_handle)

    def _set_color(self, code):
        import ctypes

        # Constants from the Windows API
        self.STD_OUTPUT_HANDLE = -11
        hdl = ctypes.windll.kernel32.GetStdHandle(self.STD_OUTPUT_HANDLE)
        ctypes.windll.kernel32.SetConsoleTextAttribute(hdl, code)

    setattr(logging.StreamHandler, '_set_color', _set_color)

    def new(*args):
        FOREGROUND_BLUE = 0x0001  # text color contains blue.
        FOREGROUND_GREEN = 0x0002  # text color contains green.
        FOREGROUND_RED = 0x0004  # text color contains red.
        FOREGROUND_INTENSITY = 0x0008  # text color is intensified.
        FOREGROUND_WHITE = FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED
        # winbase.h
        # STD_INPUT_HANDLE = -10
        # STD_OUTPUT_HANDLE = -11
        # STD_ERROR_HANDLE = -12

        # wincon.h
        # FOREGROUND_BLACK = 0x0000
        FOREGROUND_BLUE = 0x0001
        FOREGROUND_GREEN = 0x0002
        # FOREGROUND_CYAN = 0x0003
        FOREGROUND_RED = 0x0004
        FOREGROUND_MAGENTA = 0x0005
        FOREGROUND_YELLOW = 0x0006
        # FOREGROUND_GREY = 0x0007
        FOREGROUND_INTENSITY = 0x0008  # foreground color is intensified.

        # BACKGROUND_BLACK = 0x0000
        # BACKGROUND_BLUE = 0x0010
        # BACKGROUND_GREEN = 0x0020
        # BACKGROUND_CYAN = 0x0030
        # BACKGROUND_RED = 0x0040
        # BACKGROUND_MAGENTA = 0x0050
        BACKGROUND_YELLOW = 0x0060
        # BACKGROUND_GREY = 0x0070
        BACKGROUND_INTENSITY = 0x0080  # background color is intensified.

        levelno = args[1].levelno
        if (levelno >= 50):
            color = BACKGROUND_YELLOW | FOREGROUND_RED | FOREGROUND_INTENSITY | BACKGROUND_INTENSITY
        elif (levelno >= 40):
            color = FOREGROUND_RED | FOREGROUND_INTENSITY
        elif (levelno >= 30):
            color = FOREGROUND_YELLOW | FOREGROUND_INTENSITY
        elif (levelno >= 20):
            color = FOREGROUND_GREEN
        elif (levelno >= 10):
            color = FOREGROUND_MAGENTA
        else:
            color = FOREGROUND_WHITE
        args[0]._set_color(color)

        ret = fn(*args)
        args[0]._set_color(FOREGROUND_WHITE)
        # print "after"
        return ret
    return new


def add_coloring_to_emit_ansi(fn):
    # add methods we need to the class
    def new(*args):
        levelno = args[1].levelno if not isinstance(args[1], Exception) else 50
        msg = str(args[1]) if isinstance(args[1], Exception) else str(args[1].msg)

        if (levelno >= 50):
            color = '\x1b[31m'  # red
        elif (levelno >= 40):
            color = '\x1b[31m'  # red
        elif (levelno >= 30):
            color = '\x1b[33m'  # yellow
        elif (levelno >= 20):
            color = '\x1b[32m'  # green
        elif (levelno >= 10):
            color = '\x1b[35m'  # pink
        else:
            color = '\x1b[0m'  # normal

        args[1].msg = color + msg + '\x1b[0m'  # normal

        return fn(*args)
    
    return new


if platform.system() == 'Windows':
    # Windows does not support ANSI escapes and we are using API calls to set the console color
    logging.StreamHandler.emit = add_coloring_to_emit_windows(logging.StreamHandler.emit)
else:
    # all non-Windows platforms are supporting ANSI escapes so we use them
    logging.StreamHandler.emit = add_coloring_to_emit_ansi(logging.StreamHandler.emit)

CRITICAL = logging.CRITICAL
ERROR = logging.ERROR
WARNING = logging.WARNING
INFO = logging.INFO
DEBUG = logging.DEBUG
NOTSET = logging.NOTSET

def get_loglevel():
    log_level = "INFO"
    # Get from config

    # Get from environment
    if "LOG_LEVEL" in os.environ:
        log_level = os.environ["LOG_LEVEL"]
    if log_level not in ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]:
        raise Exception(f"Invalid log_level={log_level}")
    log_level = getattr(logging, log_level.upper())
    return log_level

global_log_level = None

def configure_loglevel():
    global global_log_level
    global_log_level =get_loglevel()

def getLogger(name=None, level=None):
    global global_log_level
    if not name:
        name = __name__
    logger = logging.getLogger(name)
    if not global_log_level:
        configure_loglevel()
    if not level:
        level = global_log_level
    logger.setLevel(level)
    return logger




import base64
import imghdr

def read_to_base64(image_path):
    with open(image_path, "rb") as image_file:
        data = image_file.read()
        encoded_string = base64.b64encode(data)
        base64_image =  encoded_string.decode('utf-8')
        return base64_image

def read_to_base64_imageurl(image_path):
    with open(image_path, "rb") as image_file:
        data = image_file.read()
        encoded_string = base64.b64encode(data)
        base64_image =  encoded_string.decode('utf-8')
        type = imghdr.what(None,data)
        image_url = {
            "url":f"data:image/{type};base64,{base64_image}"
        }
        return image_url

def base64_to_imageurl(base64_image):
    img_data = base64.b64decode(base64_image)
    img_type = imghdr.what(None, img_data)
    image_url = {
        "url":f"data:image/{img_type};base64,{base64_image}"
    }
    return image_url

def bytes_to_imageurl(img_data):
    img_type = imghdr.what(None, img_data)
    base64_image = base64.b64encode(img_data).decode('utf-8')
    image_url = {
        "url":f"data:image/{img_type};base64,{base64_image}"
    }
    return image_url


from gai.common import file_utils
from gai.common.logging import getLogger
logger = getLogger(__name__)
import os
from nltk.tokenize import sent_tokenize

class Chunker:

    @staticmethod
    def split(text, chunk_size=None, chunk_overlap=None):
        chunk_hash = file_utils.create_chunk_id(text)
        chunks=[]
        try:
            if chunk_size is None:
                chunk_size = 2000
            if chunk_overlap is None:
                chunk_overlap = 200
            chunks_dir = file_utils.get_chunk_dir("/tmp",chunk_hash)
            file_utils.split_chunks(text=text, 
                chunks_dir=chunks_dir,
                chunk_size=chunk_size,
                chunk_overlap=chunk_overlap)
            chunks = os.listdir(chunks_dir)
            return chunks
        except Exception as error:
            logger.error(f"Chunker.split: Failed to split chunks. error={error}")
            raise error

    @staticmethod
    def sentences(text):
        return sent_tokenize(text)




import subprocess
import os
from io import BytesIO


def play_audio(response):
    ffplay_cmd = ["ffplay", "-nodisp", "-autoexit", "-"]
    ffplay_proc = subprocess.Popen(
        ffplay_cmd, stdin=subprocess.PIPE, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)

    # Check if response is bytes
    if isinstance(response, bytes):
        # Convert bytes to BytesIO
        response = BytesIO(response)
        for chunk in iter(lambda: response.read(1024), b''):
            if chunk:
                ffplay_proc.stdin.write(chunk)
    else:
        for chunk in response.raw.stream(1024, decode_content=False):
            if chunk:
                ffplay_proc.stdin.write(chunk)

    # close on finish
    ffplay_proc.stdin.close()
    ffplay_proc.wait()


def save_audio(response, file_path):
    # Check if response is bytes
    if isinstance(response, bytes):
        with open(file_path, 'wb') as out_file:
            out_file.write(response)
    else:
        with open(file_path, 'wb') as out_file:
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    out_file.write(chunk)

    print(f"Audio saved at {os.path.abspath(file_path)}")



import os
import json
from os.path import dirname
import shutil
import sys
from dotenv import load_dotenv
load_dotenv(os.path.join(sys.path[0], '.env'))
import yaml


def init():
    with open(os.path.expanduser("~/.gairc"), "w") as file:
        file.write(json.dumps({
            "app_dir": "~/gai",
            "discovery_url": "",
        }, indent=4))
    dir_path = os.path.expanduser("~/gai")
    os.makedirs(dir_path, exist_ok=True)
    os.makedirs(os.path.join(dir_path, 'cache'), exist_ok=True)


def get_rc():
    if (not os.path.exists(os.path.expanduser("~/.gairc"))):
        init()
    return json.load(open(os.path.expanduser("~/.gairc")))


def get_api_baseurl():
    # https://gaiaio.ai/api/gen
    return os.environ.get("GAI_API_BASEURL")

def get_lib_config(config_path=None):
    rc = get_rc()
    app_dir = rc["app_dir"]
    yml = os.path.join(os.path.expanduser(app_dir), "gai.yml")
    if config_path:
        yml = config_path
    lib_config = yaml.safe_load(open(yml))
    return lib_config

def get_default_generator():
    lib_config = get_lib_config()
    return lib_config["default_generator"]


def get_generator_url(generator):
    lib_config_json = get_lib_config()
    if generator not in lib_config_json['generators']:
        raise Exception(f"Generator {generator} not found supported.")
    return lib_config_json['generators'][generator]["url"]



from gai.common.http_utils import http_post
import gai.common.ConfigHelper as ConfigHelper
from gai.lib.ClientBase import ClientBase
from gai.common.logging import getLogger
logger = getLogger(__name__)


class TTSClient(ClientBase):

    def __init__(self, config_path=None):
        super().__init__(config_path)
        logger.debug(f'base_url={self.base_url}')

    def __call__(self, input, generator="xtts-2", stream=True, **generator_params):
        if generator == "openai-tts-1":
            return self.openai_tts(input, **generator_params)

        if not input:
            raise Exception("The parameter 'input' is required.")
        data = {
            "model": generator,
            "input": input,
            "stream": stream,
            **generator_params
        }
        response = http_post(self._gen_url(generator=generator), data)
        return response

    def openai_tts(self, input, **generator_params):
        import os
        import openai
        from openai import OpenAI
        from dotenv import load_dotenv
        load_dotenv()
        if not os.environ.get("OPENAI_API_KEY"):
            raise Exception(
                "OPENAI_API_KEY not found in environment variables")
        openai.api_key = os.environ["OPENAI_API_KEY"]
        client = OpenAI()

        if not input:
            raise Exception("Missing input parameter")

        voice = generator_params.pop("voice", None)
        if not voice:
            voice = "alloy"

        generator_params.pop("language", None)
        generator_params.pop("stream", None)

        response = client.audio.speech.create(
            model='tts-1', input=input, voice=voice, **generator_params)
        return response.content



from setuptools import setup, find_packages
from os.path import abspath
import subprocess, os, sys
from setuptools.command.install import install

base_dir = os.path.dirname(os.path.abspath(__file__))
version_file = os.path.join(base_dir, 'VERSION')
with open(version_file, 'r') as f:
    VERSION = f.read().strip()

def parse_requirements(filename):
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), filename)) as f:
        required = f.read().splitlines()
    return required

setup(
    name='gai-lib',
    version=VERSION,
    author="kakkoii1337",
    author_email="kakkoii1337@gmail.com",
    packages=find_packages(),
    description = """""",
    long_description="Refer to https://gai-labs.github.io/gai/gai-lib for more information",
    long_description_content_type="text/markdown",
    classifiers=[
        'Programming Language :: Python :: 3.10',
        "Development Status :: 3 - Alpha",        
        'License :: OSI Approved :: MIT License',
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",        
        'Operating System :: OS Independent',
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules",        
        "Topic :: Scientific/Engineering :: Artificial Intelligence",        
    ],
    python_requires='>=3.10',        
    install_requires=[
        parse_requirements("requirements.txt")
    ],
    extras_require={
    },
    entry_points={
        'console_scripts': [
            'ttt=gai.cli.ttt:main',
            'tts=gai.cli.tts:main',
            'chunker=gai.cli.chunker:main',
            'pdf2txt=gai.cli.pdf2txt:main',
            'txt2md=gai.cli.txt2md:main',
            'gg=gai.cli.gg2:main',
            'summary=gai.cli.summarize:main',
            'scrape=gai.cli.scrape:main',
            'gai=gai.cli.Gaicli:main',
        ],
    }
)


import os
import gai.common.ConfigHelper as ConfigHelper

class ClientBase:

    def __init__(self, config_path=None):
        if config_path:
            self.config = ConfigHelper.get_lib_config(config_path)
        else:
            self.config = ConfigHelper.get_lib_config()
        self.base_url = self.config["gai_url"]

    def _gen_url(self, generator):
        url = os.path.join(self.base_url,
                           self.config["generators"][generator]["url"].lstrip('/'))
        return url



from fastapi import WebSocketDisconnect, WebSocket
from fastapi.websockets import WebSocketState
import asyncio
from gai.common.logging import getLogger
logger = getLogger(__name__)
import json

'''
This class carries the websocket that bridges the RAG's output
to the API server.

On one end, it is connected to the output from RAG.index_async() 
which publishes the number of chunks processed via update_progress().

On the other end, it is connected to status_update_router API which 
broadcasts the status to clients connected to '/ws'.
'''

class StatusUpdater:

    def __init__(self):
        self.status = None
        self.websocket = None

    async def connect(self, websocket: WebSocket):
        self.websocket = websocket

    async def disconnect(self, websocket: WebSocket):
        await websocket.close()
        self.websocket = None

    # update_progress is the same as update_status,
    # but it returns an integer between 0 to 100
    async def update_progress(self, i, max):
        self.status = int(i*100/max)
        if self.websocket is not None:
            if self.websocket.client_state == WebSocketState.DISCONNECTED:
                logger.info("StatusUpdater: websocket is disconnected.")
                return
            logger.info(f"StatusUpdater: sending progress {self.status}")
            await asyncio.create_task(self.websocket.send_json({'progress': self.status}))

    async def update_stop(self):
        if self.websocket is not None:
            if self.websocket.client_state == WebSocketState.DISCONNECTED:
                logger.info("StatusUpdater: websocket is disconnected.")
                return
            logger.info(f"StatusUpdater: sending <stop>")
            await asyncio.create_task(self.websocket.send_text('<stop>'))

    def get_status(self):
        return self.status



import asyncio
from gai.common.errors import ApiException
from urllib.parse import urlparse
import os
import pprint
import re
import httpx
import requests
import json
from gai.common.logging import getLogger
logger = getLogger(__name__)


def is_url(s):
    return re.match(r'^https?:\/\/.*[\r\n]*', s) is not None

# Check if URL contains a file extension (e.g. .pdf, .jpg, .png, etc.)

def has_extension(url):
    parsed_url = urlparse(url)
    _, ext = os.path.splitext(parsed_url.path)
    return bool(ext)

import json

def _handle_failed_response(response):
    error_code = "unknown"
    if response.status_code == 401:
        raise ApiException(status_code=401, code=error_code, message="Unauthorized", url=response.url)

    content_type = response.headers.get("Content-Type")
    if content_type and "application/json" in content_type:
        error_data = response.json()
    else:
        error_data = response.text()

    e = Exception()
    e.response = response
    e.status = response.status_code

    if isinstance(error_data, str):
        raise ApiException(status_code=response.status_code, code=error_code, message=error_data, url=response.url)
    
    if 'detail' in error_data:

        if isinstance(error_data['detail'], str):
            raise ApiException(status_code=response.status_code, code=error_code, message=error_data['detail'], url=response.url)

        if 'code' in error_data['detail']:
            error_code = error_data['detail']['code']

        if 'message' in error_data['detail'] and isinstance(error_data['detail']['message'], str):
            raise ApiException(status_code=response.status_code, code=error_code, message=error_data['detail']['message'], url=response.url)

    if 'code' in error_data:
        error_code = error_data['code']
        if 'message' in error_data:
            raise ApiException(status_code=response.status_code, code=error_code, message=error_data['message'], url=response.url)    
        
    raise ApiException(status_code=response.status_code, code=error_code, message=json.dumps(error_data), url=response.url)    

async def _handle_failed_response_async(response):
    error_code = "unknown"
    if response.status == 401:
        raise ApiException(status_code=401, code=error_code, message="Unauthorized", url=response.url)

    content_type = response.headers.get("Content-Type")
    if content_type and "application/json" in content_type:
        error_data = await response.json()
    else:
        error_data = await response.text()

    e = Exception()
    e.response = response
    e.status = response.status

    if isinstance(error_data, str):
        raise ApiException(status_code=response.status, code=error_code, message=error_data, url=response.url)

    if 'detail' in error_data:

        if isinstance(error_data['detail'], str):
            raise ApiException(status_code=response.status_code, code=error_code, message=error_data['detail'], url=response.url)

        if 'code' in error_data['detail']:
            error_code = error_data['detail']['code']

        if 'message' in error_data['detail'] and isinstance(error_data['detail']['message'], str):
            raise ApiException(status_code=response.status_code, code=error_code, message=error_data['detail']['message'], url=response.url)

    if 'code' in error_data:
        error_code = error_data['code']
        if 'message' in error_data:
            raise ApiException(status_code=response.status_code, code=error_code, message=error_data['message'], url=response.url)    
        
    raise ApiException(status_code=response.status_code, code=error_code, message=json.dumps(error_data), url=response.url)    

async def http_post_async(url, data):
    return httppost_async(url, data)

async def httppost_async(url, data):
    headers = {"Content-Type": "application/json"}
    async with httpx.AsyncClient() as client:
        if not isinstance(data, str):
            data = json.dumps(data)
        try:
            response = await client.post(url, data=data, headers=headers)
            if response.status == 200:
                return response
            else:
                await _handle_failed_response_async(response)
        except httpx.HTTPStatusError as e:
            raise Exception("Connection Error. Is the service Running?")


def http_post(url, data=None, files=None):
    return httppost(url, data, files)


def httppost(url, data=None, files=None):
    if data == None and files == None:
        raise Exception("No data or files provided")

    logger.debug(f"httppost:url={url}")
    logger.debug(f"httppost:data={pprint.pformat(data)}")
    try:
        if files:
            if data and "stream" in data:
                files["stream"] = (None, data["stream"])
            response = requests.post(url, files=files)
        else:
            if "stream" in data:
                response = requests.post(url, json=data, stream=data["stream"])
            else:
                response = requests.post(url, json=data)
        if response.status_code == 200:
            return response
        else:
            _handle_failed_response(response)

    except requests.exceptions.ConnectionError as e:
        raise Exception("Connection Error. Is the service Running?")


def http_get(url):
    return httpget(url)

def httpget(url):
    try:
        response = requests.get(url)
        if response.status_code == 200:
            return response
        else:
            _handle_failed_response(response)
    except requests.exceptions.ConnectionError as e:
        raise Exception("Connection Error. Is the service Running?")

async def http_get_async(url):
    return httpget_async(url)

async def httpget_async(url):
    async with httpx.AsyncClient() as session:
        try:
            async with session.get(url) as response:
                if response.status == 200:
                    return response                      # Returning the data
                else:
                    await _handle_failed_response_async(response)
        except httpx.HTTPStatusError as e:
            raise Exception("Connection Error. Is the service Running?")

### DELETE method

def http_delete(url):
    return httpdelete(url)

def httpdelete(url):
    try:
        response = requests.delete(url)
        if response.status_code == 200:
            return response
        else:
            _handle_failed_response(response)
    except requests.exceptions.ConnectionError as e:
        raise Exception("Connection Error. Is the service Running?")

async def http_delete_async(url):
    return httpdelete_async(url)

async def httpdelete_async(url):
    async with httpx.AsyncClient() as session:
        try:
            response = await session.delete(url)
            if response.status_code == 200:
                return response
            else:
                await _handle_failed_response_async(response)
        except httpx.HTTPStatusError as e:
            raise Exception("Connection Error. Is the service Running?")

### PUT method

def http_put(url):
    return httpput(url)

def httpput(url):
    try:
        response = requests.put(url)
        if response.status_code == 200:
            return response
        else:
            _handle_failed_response(response)
    except requests.exceptions.ConnectionError as e:
        raise Exception("Connection Error. Is the service Running?")

async def http_put_async(url):
    return httpput_async(url)

async def httpput_async(url):
    async with httpx.AsyncClient() as session:
        try:
            response = await session.put(url)
            if response.status_code == 200:
                return response
            else:
                await _handle_failed_response_async(response)
        except httpx.HTTPStatusError as e:
            raise Exception("Connection Error. Is the service Running?")
